//********************************************************************************************************************************
//*                                                                                                                              *
//*     TSpectrum3D 1.1.8.12  3delite 2008-2013                                                                                 *
//*     See TSpectrum3D ReadMe.txt for details                                                                                   *
//*                                                                                                                              *
//* Two licenses are available if you use this library:                                                                          *
//* Shareware License: 50 Euros                                                                                                  *
//* Commercial License: 250 Euros                                                                                                *
//*                                                                                                                              *
//*     http://www.3delite.hu/Object Pascal Developer Resources/tspectrum3d.html                                                 *
//*                                                                                                                              *
//* Note: If you plan using TSpectrum3D in you programs (shareware or commercial), you'll also need a BASS license               *
//* available separatelly from:                                                                                                  *
//*                                                                                                                              *
//*     http://www.un4seen.com/bass.html                                                                                         *
//*                                                                                                                              *
//*     Credits:                                                                                                                 *
//* Ian Luck @ www.un4seen.com - for the great BASS audio library.                                                               *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit Spectrum3D;

//* Remove the "." from the following line to build with BASSWASAPI support
{.$DEFINE USEBASSWASAPI}

//* Default is Clootie DX9 units
{$DEFINE CLOOTIEDXUNITS}

//* Targeting Win64 only supports Delphi XE2's DX9 units
//* Note that on target system DirectX9 needs explicitly to be installed
{$IFDEF WIN64}
    {$DEFINE DELPHIDXUNITS}
    {$UNDEF CLOOTIEDXUNITS}
{$ENDIF}

//* Remove the "." from the following line to force building with Delphi XE2 DX9 units
{.$DEFINE DELPHIDXUNITS}

{$IFDEF DELPHIDXUNITS}
    {$UNDEF CLOOTIEDXUNITS}
{$ENDIF}

{$IFDEF CLOOTIEDXUNITS}
    {$IFDEF WIN64}
    Building with Clootie DX9 package only supports the 32 bit platform!
    {$ENDIF}
{$ENDIF}

interface

uses
    Classes,
    Controls,
    Graphics,
    Windows,
    Forms,
    Messages,
    StdCtrls,
    ExtCtrls,
    SysUtils,
    Dialogs,
    {$IFDEF DELPHIDXUNITS}
    //* Delphi XE2 or higher
    Winapi.Direct3D9,
    Winapi.d3dx9,
    Winapi.DXTypes,
    {$ELSE}
    //* Clootie package
    Direct3D9,
    d3dx9,
    DXTypes,
    {$ENDIF}
    {$IFDEF USEBASSWASAPI}
    BASSWASAPI,
    {$ENDIF}
    BASS;


type
    TSpectrum3DBeforeD3DRender = procedure(Sender: TObject; Index: Cardinal) of object;
    TSpectrum3DAfterD3DRender = procedure(Sender: TObject; Index: Cardinal) of object;
    TSpectrum3DOnD3DDestroy = procedure(Sender: TObject) of object;
    TSpectrum3DOnMouse = procedure(Sender: TObject; Button: TMouseButton; X, Y: Integer) of object;
    TSpectrum3DOnMouseMove = procedure(Sender: TObject; X, Y: Integer) of object;
    TSpectrum3DOnMouseWheel = procedure(Sender: TObject; var WheelDelta: SmallInt) of object;
    TSpectrum3DOnKeyDown = procedure(Sender: TObject; var CharCode: Word) of object;
    TSpectrum3DOnDblClick = procedure(Sender: TObject; X, Y: Integer) of object;

type
    TD3DVertexProcessingMode = (vpSoftware, vpHardware, vpMixed);

    //* Helper functions
    function BGR2RGB(BGRValue: Cardinal): Cardinal;

//* The picture class
type
    TSamplePicture = class(TImage)
    private
        procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
        procedure WMMouseWheel(var Message: TWMMouseWheel); message WM_MOUSEWHEEL;
        procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
        procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
        procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
        procedure WMLButtonDblClick(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
        procedure WMRButtonDown(var Message: TWMRButtonDown); message WM_RBUTTONDOWN;
        procedure WMRButtonUp(var Message: TWMRButtonUp); message WM_RBUTTONUP;
        procedure WMMButtonDown(var Message: TWMMButtonDown); message WM_MBUTTONDOWN;
    protected
    public
        Constructor Create(AOwner: TComponent); override;
        Destructor Destroy; override;
    published
    end;

//* The panel that will be used for rendering
type
    TPanelSpectrum = class(TPanel)
    private
        FPicture: TSamplePicture;
        procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
        procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
        procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
        procedure WMLButtonDblClick(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
        procedure WMRButtonDown(var Message: TWMRButtonDown); message WM_RBUTTONDOWN;
        procedure WMRButtonUp(var Message: TWMRButtonUp); message WM_RBUTTONUP;
        procedure WMMButtonDown(var Message: TWMMButtonDown); message WM_MBUTTONDOWN;
        procedure WMMouseWheel(var Message: TWMMouseWheel); message WM_MOUSEWHEEL;
    protected
    public
        Constructor Create(AOwner: TComponent); override;
        Destructor Destroy; override;
    published
        property OnExit;
        property Handle;
    end;

//* The panel that holds everything together
type
    [ComponentPlatformsAttribute(pidWin32 or pidWin64)]
    TSpectrum3D = class(TPanel)
    private
        dBackColor: TColor;
        dwFrames: Cardinal;
        dFPS: Int64;
        FBeforeD3DRender: TSpectrum3DBeforeD3DRender;
        FAfterD3DRender: TSpectrum3DAfterD3DRender;
        FOnD3DDestroy: TSpectrum3DOnD3DDestroy;
        FOnMouseDown: TSpectrum3DOnMouse;
        FOnMouseUp: TSpectrum3DOnMouse;
        FOnMouseMove: TSpectrum3DOnMouseMove;
        FOnMouseWheel: TSpectrum3DOnMouseWheel;
        FOnKeyDown: TSpectrum3DOnKeyDown;
        bSuppressErrors: Boolean;
        dVertexProcessingMode: TD3DVertexProcessingMode;
        FOnDblClick: TSpectrum3DOnDblClick;
        VertexBuffer_Grid: IDirect3DVertexBuffer9;
        VertexBuffer_FFT: IDirect3DVertexBuffer9;
        VertexBuffer_Font: ID3DXFont;
        matWorld: TD3DMatrix;
        matRotSample: TD3DMatrix;
        SampleAmplitude: Single;
        Rendering: Boolean;
        FAntialiasing: Boolean;
        FShowText: Boolean;
        FShowGridLines: Boolean;
        PFFTVertexes: Pointer;
        FGradientGrid: Boolean;
        procedure OnPanelResize(Sender: TObject);
        procedure SetPanel3DColor(const Color: TColor);
        function GetPanel3DColor: TColor;
        procedure SetPicture(const Pic: TPicture);
        function GetPicture: TPicture;
        procedure SetPictureVisibility(const Visible: Boolean);
        function GetPictureVisibility: Boolean;
        procedure WMMouseWheel(var Message: TWMMouseWheel); message WM_MOUSEWHEEL;
        function InitGridGeometry(d3dDevice: IDirect3DDevice9; var d3dVertexBuffer: IDirect3DVertexBuffer9): HRESULT;
        function InitGridGeometrySolid(d3dDevice: IDirect3DDevice9; var d3dVertexBuffer: IDirect3DVertexBuffer9): HRESULT;
        function InitGeometry(d3dDevice: IDirect3DDevice9; var d3dVertexBuffer: IDirect3DVertexBuffer9; Channel: Cardinal): HRESULT;
        procedure SetupCameraMatrix(d3dDevice: IDirect3DDevice9);
    public
        PanelSpectrum: TPanelSpectrum;
        g_pD3D: IDirect3D9;
        g_pd3dDevice: IDirect3DDevice9;
        GFXZoom: Single;
        GFXScroll: Single;
        GFXAmplitude: Single;
        Channel: DWord;
        FFTSize: Cardinal;
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        property FPS: Int64 read dFPS;
    published
        procedure Close;
        function InitD3D: Boolean;
        procedure D3DFree;
        procedure Render;
        property BeforeD3DRender: TSpectrum3DBeforeD3DRender read FBeforeD3DRender write FBeforeD3DRender;
        property AfterD3DRender: TSpectrum3DAfterD3DRender read FAfterD3DRender write FAfterD3DRender;
        property SuppressErrors: Boolean read bSuppressErrors write bSuppressErrors default False;
        property Panel3DColor: TColor read GetPanel3DColor write SetPanel3DColor;
        property Panel3DPicture: TPicture read GetPicture write SetPicture;
        property Panel3DPictureVisible: Boolean read GetPictureVisibility write SetPictureVisibility default True;
        property OnD3DDestroy: TSpectrum3DOnD3DDestroy read FOnD3DDestroy write FOnD3DDestroy;
        property OnMouseDown: TSpectrum3DOnMouse read FOnMouseDown write FOnMouseDown;
        property OnMouseUp: TSpectrum3DOnMouse read FOnMouseUp write FOnMouseUp;
        property OnMouseMove: TSpectrum3DOnMouseMove read FOnMouseMove write FOnMouseMove;
        property OnMouseWheel: TSpectrum3DOnMouseWheel read FOnMouseWheel write FOnMouseWheel;
        property OnKeyDown: TSpectrum3DOnKeyDown read FOnKeyDown write FOnKeyDown;
        property SpectrumBackColor: TColor read dBackColor write dBackColor;
        property D3DVertexProcessingMode: TD3DVertexProcessingMode read dVertexProcessingMode write dVertexProcessingMode default vpHardware;
        property OnDblClick: TSpectrum3DOnDblClick read FOnDblClick write FOnDblClick;
        property Antialiasing: Boolean read FAntialiasing write FAntialiasing default False;
        property ShowText: Boolean read FShowText write FShowText default True;
        property ShowGridLines: Boolean read FShowGridLines write FShowGridLines default True;
        property GradientGrid: Boolean read FGradientGrid write FGradientGrid default False;
    end;

	procedure Register;

type
    PCustomVertex = ^TCustomVertex;
    TCustomVertex = record
        x, y, z: Single;            // The untransformed, 3D position for the vertex
        Color: DWORD;               // The vertex color
    end;

const
    D3DFVF_CUSTOMVERTEX = (D3DFVF_XYZ or D3DFVF_DIFFUSE);
    D3D_Y_AMPLITUDE = 4000;
    //SIZE = 1727;
    SIZE = 1860;
    GRID_VERTEX_COUNT = 26;

implementation

procedure Register;
begin
    RegisterComponents('3delite', [TSpectrum3D]);
end;

function D3DXVector3(_x, _y, _z: Single): TD3DXVector3;
begin
    with Result do begin
        x:= _x;
        y:= _y;
        z:=_z;
    end;
end;

function BGR2RGB(BGRValue: Cardinal): Cardinal;
begin
    BGRValue := (BGRValue shl 8) shr 8;
   	Result := (BGRValue shr 16) + (((BGRValue shl 16) shr 24) shl 8) + (BGRValue shl 16);
    Result := $FF000000 OR Result;
end;

constructor TPanelSpectrum.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    FPicture := TSamplePicture.Create(Self);
    FPicture.Parent := Self;
    FPicture.Align := alClient;
    FPicture.Stretch := True;
    DoubleBuffered := True;
end;

destructor TPanelSpectrum.Destroy;
begin
    FPicture.Free;
    inherited Destroy;
end;

procedure TPanelSpectrum.WMMouseMove(var Message: TWMMouseMove);
begin
    try
        with (Parent as TSpectrum3D) do begin
            if Assigned(FOnMouseMove)
                then FOnMouseMove(Parent, Message.XPos, Message.YPos);
        end;
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TPanelSpectrum.WMLButtonDown(var Message: TWMLButtonDown);
begin
    try
        with (Parent as TSpectrum3D) do begin
            if Assigned(FOnMouseDown)
                then FOnMouseDown(Parent, mbLeft, Message.XPos, Message.YPos);
        end;
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TPanelSpectrum.WMLButtonUp(var Message: TWMLButtonUp);
begin
    try
         if Assigned((Parent as TSpectrum3D).FOnMouseUp)
            then (Parent as TSpectrum3D).FOnMouseUp(Parent, mbLeft, Message.XPos, Message.YPos);
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TPanelSpectrum.WMLButtonDblClick(var Message: TWMLButtonDblClk);
begin
    try
        if Assigned((Parent as TSpectrum3D).FOnDblClick)
            then (Parent as TSpectrum3D).FOnDblClick(Parent, Message.XPos, Message.YPos);
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TPanelSpectrum.WMRButtonDown(var Message: TWMRButtonDown);
begin
    try
        with (Parent as TSpectrum3D) do begin
            if Assigned(FOnMouseDown)
                then FOnMouseDown(Parent, mbRight, Message.XPos, Message.YPos);
        end;
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TPanelSpectrum.WMRButtonUp(var Message: TWMRButtonUp);
begin
    try
        if Assigned((Parent as TSpectrum3D).FOnMouseUp)
            then (Parent as TSpectrum3D).FOnMouseUp(Parent, mbRight, Message.XPos, Message.YPos);
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TPanelSpectrum.WMMButtonDown(var Message: TWMMButtonDown);
begin
    try
        with (Parent as TSpectrum3D) do begin
            if Assigned(FOnMouseDown)
                then FOnMouseDown(Parent, mbMiddle, Message.XPos, Message.YPos);
        end;
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TPanelSpectrum.WMMouseWheel(var Message: TWMMouseWheel);
begin
    try
        with (Parent as TSpectrum3D) do begin
            if Assigned(FOnMouseWheel)
                then FOnMouseWheel(Parent, Message.WheelDelta);
        end;
        DefaultHandler(Message);
    except
        //*
    end;
end;

constructor TSpectrum3D.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    DoubleBuffered := True;
    PanelSpectrum := TPanelSpectrum.Create(Self);
    PanelSpectrum.Parent := Self;
    PanelSpectrum.Align := alClient;
    PanelSpectrum.Caption := 'TSpectrum3D by 3delite';
    PanelSpectrum.BevelOuter := bvNone;
    PanelSpectrum.OnResize := OnPanelResize;
    PanelSpectrum.Color := Color;
    PanelSpectrum.FPicture.Visible := True;
    dBackColor := $000000;
    dVertexProcessingMode := vpHardware;
    BevelOuter := bvLowered;
    bSuppressErrors := False;
    g_pD3D := nil;
    g_pd3dDevice := nil;
    GFXZoom := 101;
    GFXScroll := 3;
    GFXAmplitude := 1.0;
    FAntialiasing := False;
    FShowText := True;
    FShowGridLines := True;
    FGradientGrid := False;
    PFFTVertexes := nil;
    FFTSize := BASS_DATA_FFT4096;
end;

destructor TSpectrum3D.Destroy;
begin
    D3DFree;
    PanelSpectrum.Free;
    inherited Destroy;
end;

function TSpectrum3D.InitD3D: Boolean;
var
    d3dpp: TD3DPresentParameters;
    InitResult: HResult;
    Flags: Cardinal;
begin
    Result := False;
    try
        if g_pD3D = nil then begin
            try
                g_pD3D := Direct3DCreate9(D3D_SDK_VERSION);
                if g_pD3D = nil then begin
                    if NOT bSuppressErrors
                        then MessageDlg('Error: Creating D3D Object!', mtWarning, [mbOk], 0);
                    Exit;
                end;
            except
                //*
            end;
        end;
        FillChar(d3dpp, SizeOf(d3dpp), 0);
        d3dpp.Windowed := True;
        d3dpp.SwapEffect := D3DSWAPEFFECT_DISCARD;
        d3dpp.BackBufferFormat := D3DFMT_UNKNOWN;
        g_pd3dDevice := nil;
        case dVertexProcessingMode of
            vpSoftware: Flags := D3DCREATE_SOFTWARE_VERTEXPROCESSING;
            vpHardware: Flags := D3DCREATE_HARDWARE_VERTEXPROCESSING;
            vpMixed:    Flags := D3DCREATE_MIXED_VERTEXPROCESSING;
        else Flags := D3DCREATE_HARDWARE_VERTEXPROCESSING;
        end;
        //* SysUtils.Now bug workaround
        Flags := Flags OR D3DCREATE_FPU_PRESERVE;
        InitResult := g_pD3D.CreateDevice(D3DADAPTER_DEFAULT,
            D3DDEVTYPE_HAL,
            PanelSpectrum.Handle,
            Flags OR D3DCREATE_MULTITHREADED,
            @d3dpp,
            g_pd3dDevice);
        if FAILED(InitResult) then begin
            if NOT bSuppressErrors
                then MessageDlg('Error: Creating D3D Device!', mtWarning, [mbOk], 0);
            Result := False;
            Exit;
        end;
        g_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
        if FAntialiasing then begin
            g_pd3dDevice.SetRenderState(D3DRS_MULTISAMPLEANTIALIAS, iTrue);
        end;
        g_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_ANISOTROPIC);
        g_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_ANISOTROPIC);
        g_pd3dDevice.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_ANISOTROPIC);
        g_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE, 1);
        g_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iFalse);
        if FAntialiasing then begin
            g_pd3dDevice.SetRenderState(D3DRS_ANTIALIASEDLINEENABLE, iTrue);
        end;
        D3DXMatrixIdentity(matRotSample);
        D3DXMatrixRotationX(matRotSample, Pi / 2);
        try
            GetMem(PFFTVertexes, SIZE * SizeOf(TCustomVertex));
        except
            if NOT bSuppressErrors
                    then Showmessage('Error: Exception: PFFTVertexes GetMem: ' + IntToStr(SIZE * SizeOf(TCustomVertex)));
            Exit;
        end;
        if FGradientGrid then begin
            InitGridGeometry(g_pd3dDevice, VertexBuffer_Grid);
        end else begin
            InitGridGeometrySolid(g_pd3dDevice, VertexBuffer_Grid);
        end;
        InitGeometry(g_pd3dDevice, VertexBuffer_FFT, Channel);
        // Create font
        D3DXCreateFontW(g_pd3dDevice,          // D3D device
            10,               // Height
            0,                     // Width
            FW_THIN,               // Weight
            0,                     // MipLevels, 0 = autogen mipmaps
            FALSE,                 // Italic
            DEFAULT_CHARSET,       // CharSet
            OUT_DEFAULT_PRECIS,    // OutputPrecision
            DEFAULT_QUALITY,       // Quality
            DEFAULT_PITCH or FF_DONTCARE, // PitchAndFamily
            PWideChar('Verdana'),              // pFaceName
            VertexBuffer_Font);
        PanelSpectrum.FPicture.Hide;
        Result := True;
    except
        if NOT bSuppressErrors
            then MessageDlg('Error: Creating D3D Device!' + (#13#13) + 'TSpectrum3D.InitD3D()', mtWarning, [mbOk], 0);
    end;
end;

procedure TSpectrum3D.Render;
const
    FONT_COLOR = $ffdddddd;
var
    rc: TRect;
    pRect: TRect;
    Scroll: Single;
    DisplayWidth: Integer;
    ChannelInfo: BASS_CHANNELINFO;
    {$IFDEF USEBASSWASAPI}
    WASAPIInfo: BASS_WASAPI_INFO;
    {$ENDIF}
    SampleRate: Integer;
    FreqText1: AnsiString;
    FreqText2: AnsiString;
    FreqText3: AnsiString;
    FreqText4: AnsiString;
    FreqText5: AnsiString;
    FreqText6: AnsiString;
    FreqText7: AnsiString;
begin
    try
        if g_pd3dDevice = nil
            then Exit;
        Rendering := True;
        if Assigned(BeforeD3DRender)
            then BeforeD3DRender(Self, 0);
        Windows.GetClientRect(PanelSpectrum.Handle, pRect);
        DisplayWidth := pRect.Right;
        with g_pd3dDevice do begin
            // Clear the backbuffer to a black color       D3DCOLOR_XRGB(0,0,0)
            Clear(0, nil, D3DCLEAR_TARGET, BGR2RGB(dBackColor), 1.0, 0);
            if (SUCCEEDED(BeginScene)) then begin
                try
                    //if Bass_ChannelIsActive(Channel) = BASS_ACTIVE_PLAYING then begin
                        InitGeometry(g_pd3dDevice, VertexBuffer_FFT, Channel);
                    //end;
                except
                    if NOT bSuppressErrors
                        then MessageDlg('Error: Exception: Init geometry.', mtError, [mbOk], 0);
                end;
                // Setup the world, view, and projection matrices
                SetupCameraMatrix(g_pd3dDevice);
                //* Render grid
                if FShowGridLines then begin
                    SetStreamSource(0, VertexBuffer_Grid, 0, SizeOf(TCustomVertex));
                    SetFVF(D3DFVF_CUSTOMVERTEX);
                    DrawPrimitive(D3DPT_LINELIST, 0, GRID_VERTEX_COUNT div 2);
                end;
                // Render the vertex buffer contents
                SetStreamSource(0, VertexBuffer_FFT, 0, SizeOf(TCustomVertex));
                SetFVF(D3DFVF_CUSTOMVERTEX);
                DrawPrimitive(D3DPT_LINESTRIP, 0, SIZE - 2);
                if FShowText then begin
                    try

                        {$IFDEF USEBASSWASAPI}
                        if ((Channel > 0) AND (Channel < 255)) then begin
                            if BASS_WASAPI_SetDevice(Channel - 1) then begin
                                BASS_WASAPI_GetInfo(WASAPIInfo);
                                SampleRate := WASAPIInfo.freq;
                            end;
                        end else begin
                        {$ENDIF}
                            BASS_ChannelGetInfo(Channel, ChannelInfo);
                            SampleRate := ChannelInfo.freq;
                        {$IFDEF USEBASSWASAPI}
                        end;
                        {$ENDIF}

                        if SampleRate < 1 then begin
                            SampleRate := 48000;
                        end;

                        case SampleRate of
                            44100: begin
                                FreqText1 := '92Hz';
                                FreqText2 := '920Hz';
                                FreqText3 := '1.8KHz';
                                FreqText4 := '4.6KHz';
                                FreqText5 := '9.2KHz';
                                FreqText6 := '13.8KHz';
                                FreqText7 := '18.3KHz';
                            end;
                            48000: begin
                                FreqText1 := '100Hz';
                                FreqText2 := '1KHz';
                                FreqText3 := '2KHz';
                                FreqText4 := '5KHz';
                                FreqText5 := '10KHz';
                                FreqText6 := '15KHz';
                                FreqText7 := '20KHz';
                            end;
                            96000: begin
                                FreqText1 := '200Hz';
                                FreqText2 := '2KHz';
                                FreqText3 := '4KHz';
                                FreqText4 := '10KHz';
                                FreqText5 := '20KHz';
                                FreqText6 := '30KHz';
                                FreqText7 := '40KHz';
                            end;
                            192000: begin
                                FreqText1 := '400Hz';
                                FreqText2 := '4KHz';
                                FreqText3 := '8KHz';
                                FreqText4 := '20KHz';
                                FreqText5 := '40KHz';
                                FreqText6 := '60KHz';
                                FreqText7 := '80KHz';
                            end;
                            else begin
                                FreqText1 := FloatToStrF((100 / 48000) * SampleRate, ffGeneral, 8, 1) + 'Hz';
                                FreqText2 := FloatToStrF(((1000 / 48000) * SampleRate) / 1000, ffGeneral, 8, 1) + 'KHz';
                                FreqText3 := FloatToStrF(((2000 / 48000) * SampleRate) / 1000, ffGeneral, 8, 1) + 'KHz';
                                FreqText4 := FloatToStrF(((5000 / 48000) * SampleRate) / 1000, ffGeneral, 8, 1) + 'KHz';
                                FreqText5 := FloatToStrF(((10000 / 48000) * SampleRate) / 1000, ffGeneral, 8, 1) + 'KHz';
                                FreqText6 := FloatToStrF(((15000 / 48000) * SampleRate) / 1000, ffGeneral, 8, 1) + 'KHz';
                                FreqText7 := FloatToStrF(((20000 / 48000) * SampleRate) / 1000, ffGeneral, 8, 1) + 'KHz';
                            end;
                        end;

                        // Draw text   4.62   - 3.59
                        pRect.Right := Trunc(pRect.Right * (GFXZoom / 100));
                        Scroll := ((4.62 / 1188 * DisplayWidth) * (GFXZoom - 100) / (500 - 100)) * (GFXScroll);
                                                        //1.032
                        SetRect(rc, Trunc((pRect.Right / 1.06) - Scroll), pRect.Bottom - 12, 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar(FreqText7), -1, @rc, DT_NOCLIP, FONT_COLOR);
                        SetRect(rc, Trunc((pRect.Right / 1.388) - Scroll), pRect.Bottom - 12, 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar(FreqText6), -1, @rc, DT_NOCLIP, FONT_COLOR);
                        SetRect(rc, Trunc((pRect.Right / 2.08) - Scroll), pRect.Bottom - 12, 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar(FreqText5), -1, @rc, DT_NOCLIP, FONT_COLOR);
                        SetRect(rc, Trunc((pRect.Right / 4.14) - Scroll), pRect.Bottom - 12, 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar(FreqText4), -1, @rc, DT_NOCLIP, FONT_COLOR);
                        SetRect(rc, Trunc((pRect.Right / 10) - Scroll), pRect.Bottom - 12, 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar(FreqText3), -1, @rc, DT_NOCLIP, FONT_COLOR);
                        SetRect(rc, Trunc((pRect.Right / 20) - Scroll), pRect.Bottom - 12, 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar(FreqText2), -1, @rc, DT_NOCLIP, FONT_COLOR);
                        SetRect(rc, Trunc((pRect.Right / 300) - Scroll), pRect.Bottom - 12, 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar(FreqText1), -1, @rc, DT_NOCLIP, FONT_COLOR);

                        SetRect(rc, 3, Trunc(pRect.Bottom / 17), 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar('  0dB'), -1, @rc, DT_NOCLIP, FONT_COLOR);
                        SetRect(rc, 3, Trunc(pRect.Bottom / 2.14), 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar('-20dB'), -1, @rc, DT_NOCLIP, FONT_COLOR);
                        SetRect(rc, 3, Trunc(pRect.Bottom / 1.525), 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar('-40dB'), -1, @rc, DT_NOCLIP, FONT_COLOR);
                        SetRect(rc, 3, Trunc(pRect.Bottom / 1.25), 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar('-60dB'), -1, @rc, DT_NOCLIP, FONT_COLOR);
                        SetRect(rc, 3, Trunc(pRect.Bottom / 1.115), 0, 0);
                        VertexBuffer_Font.DrawTextA(nil, PANSIChar('-75dB'), -1, @rc, DT_NOCLIP, FONT_COLOR);
                    except
                        if NOT bSuppressErrors
                            then MessageDlg('Error: Exception: Rendering text.', mtError, [mbOk], 0);
                    end;
                end;
                if Assigned(AfterD3DRender)
                    then AfterD3DRender(Self, 0);
                EndScene;
            end;
            Present(nil, nil, 0, nil);
            Inc(dwFrames);
            Rendering := False;
        end;
    except
        Rendering := False;
        if NOT bSuppressErrors
            then MessageDlg('Error: Exception: Rendering frame.', mtError, [mbOk], 0);
    end;
end;

procedure TSpectrum3D.Close;
begin
    try
        //BASS_StreamFree(Channel);
        Channel := 0;
        g_pd3dDevice := nil;
        VertexBuffer_Grid := nil;
        VertexBuffer_Font := nil;
        VertexBuffer_FFT := nil;
        D3DXMatrixIdentity(matWorld);
        D3DXMatrixIdentity(matRotSample);
        SampleAmplitude := 1.0;
        Rendering := False;
        Tag := 0;
    except
        //*
    end;
    //* Check to free D3D Object (not really tested)
    D3DFree;
end;

procedure TSpectrum3D.OnPanelResize;
begin
    D3DFree;
    InitD3D;
end;

procedure TSpectrum3D.D3DFree;
begin
    if g_pD3D <> nil then begin
        if Assigned(FOnD3DDestroy)
            then FOnD3DDestroy(Self);
        g_pD3D := nil;
    end;
    if Assigned(PFFTVertexes) then begin
        FreeMem(PFFTVertexes);
        PFFTVertexes := nil;
    end;
end;

function TSpectrum3D.GetPanel3DColor: TColor;
begin
    Result := PanelSpectrum.Color;
end;

procedure TSpectrum3D.SetPanel3DColor(const Color: TColor);
begin
    PanelSpectrum.Color := Color;
end;

function TSpectrum3D.GetPicture: TPicture;
begin
    Result := PanelSpectrum.FPicture.Picture;
end;

procedure TSpectrum3D.SetPicture(const Pic : TPicture);
begin
    PanelSpectrum.FPicture.Picture := Pic;
end;

function TSpectrum3D.GetPictureVisibility: Boolean;
begin
    Result := PanelSpectrum.FPicture.Visible;
end;

procedure TSpectrum3D.SetPictureVisibility(const Visible: Boolean);
begin
    PanelSpectrum.FPicture.Visible := Visible;
end;

procedure TSpectrum3D.WMMouseWheel(var Message: TWMMouseWheel);
begin
    try
        if Assigned(FOnMouseWheel)
            then FOnMouseWheel(Self, Message.WheelDelta);
        DefaultHandler(Message);
    except
        //*
    end;
end;

{ TSamplePicture }

constructor TSamplePicture.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
end;

destructor TSamplePicture.Destroy;
begin
    inherited Destroy;
end;

procedure TSamplePicture.WMKeyDown(var Message: TWMKeyDown);
begin
    try
        if Assigned((Parent.Parent as TSpectrum3D).FOnKeyDown)
            then (Parent.Parent as TSpectrum3D).FOnKeyDown(Parent, Message.CharCode);
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TSamplePicture.WMLButtonDblClick(var Message: TWMLButtonDblClk);
begin
    try
        if Assigned((Parent.Parent as TSpectrum3D).FOnDblClick)
            then (Parent.Parent as TSpectrum3D).FOnDblClick(Parent.Parent, Message.XPos, Message.YPos);
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TSamplePicture.WMLButtonDown(var Message: TWMLButtonDown);
begin
    try
        if Assigned((Parent.Parent as TSpectrum3D).FOnMouseDown)
            then (Parent.Parent as TSpectrum3D).FOnMouseDown(Parent.Parent, mbLeft, Message.XPos, Message.YPos);
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TSamplePicture.WMLButtonUp(var Message: TWMLButtonUp);
begin
    try
        if Assigned((Parent.Parent as TSpectrum3D).FOnMouseUp)
            then (Parent.Parent as TSpectrum3D).FOnMouseUp(Parent.Parent, mbLeft, Message.XPos, Message.YPos);
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TSamplePicture.WMMButtonDown(var Message: TWMMButtonDown);
begin
    try
        if Assigned((Parent.Parent as TSpectrum3D).FOnMouseDown)
            then (Parent.Parent as TSpectrum3D).FOnMouseDown(Parent, mbMiddle, Message.XPos, Message.YPos);
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TSamplePicture.WMMouseMove(var Message: TWMMouseMove);
begin
    try
        if Assigned((Parent.Parent as TSpectrum3D).FOnMouseMove)
            then (Parent.Parent as TSpectrum3D).FOnMouseMove(Parent.Parent, Message.XPos, Message.YPos);
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TSamplePicture.WMMouseWheel(var Message: TWMMouseWheel);
begin
    try
        if Assigned((Parent.Parent as TSpectrum3D).FOnMouseWheel)
            then (Parent.Parent as TSpectrum3D).FOnMouseWheel(Parent, Message.WheelDelta);
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TSamplePicture.WMRButtonDown(var Message: TWMRButtonDown);
begin
    try
        if Assigned((Parent.Parent as TSpectrum3D).FOnMouseDown)
            then (Parent.Parent as TSpectrum3D).FOnMouseDown(Parent, mbRight, Message.XPos, Message.YPos);
        DefaultHandler(Message);
    except
        //*
    end;
end;

procedure TSamplePicture.WMRButtonUp(var Message: TWMRButtonUp);
begin
    try
        if Assigned((Parent.Parent as TSpectrum3D).FOnMouseUp)
            then (Parent.Parent as TSpectrum3D).FOnMouseUp(Parent, mbRight, Message.XPos, Message.YPos);
        DefaultHandler(Message);
    except
        //*
    end;
end;

function TSpectrum3D.InitGridGeometry(d3dDevice: IDirect3DDevice9; var d3dVertexBuffer: IDirect3DVertexBuffer9): HRESULT;
const
    GRID_COLORD = $ff303030;
    GRID_COLORL = $ff080808;
    DEPTH = 0;
    BOTTOM = 0;
    SIZE = 1727;
var
    PVertices: Pointer;
    PVertex: ^TCustomVertex;
    PVertexes: Pointer;
    VertexBufferSize: Cardinal;
    ColorGridD: Cardinal;
    ColorGridF: Cardinal;
begin
    Result:= E_FAIL;
    try
        GetMem(PVertexes, GRID_VERTEX_COUNT * SizeOf(TCustomVertex));
    except
        if NOT bSuppressErrors
            then MessageDlg('Error: GetMem: ' + IntToStr(GRID_VERTEX_COUNT * SizeOf(TCustomVertex)), mtWarning, [mbOk], 0);
        Exit;
    end;
    ColorGridD := GRID_COLORD;
    ColorGridF := $FF000000;
    PVertex := PVertexes;
    try
        PVertex.x := 0;
        PVertex.y := 0;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := SIZE * 1.05;
        PVertex.y := 0;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);

        PVertex.x := 0;
        PVertex.y := 1 - 1 / 1.083;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := SIZE * 1.05;
        PVertex.y := 1 - 1 / 1.083;
        PVertex.z := DEPTH;
        PVertex.color := GRID_COLORL;
        Inc(PVertex);

        PVertex.x := 0;
        PVertex.y := 1 - 1 / 1.225;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := SIZE * 1.05;
        PVertex.y := 1 - 1 / 1.225;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridF;
        Inc(PVertex);

        PVertex.x := 0;
        PVertex.y := 1 - 1 / 1.525;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := SIZE * 1.05;
        PVertex.y := 1 - 1 / 1.525;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridF;
        Inc(PVertex);

        PVertex.x := 0;
        PVertex.y := 1 - 1 / 2.25;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := SIZE * 0.6;
        PVertex.y := 1 - 1 / 2.25;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridF;
        Inc(PVertex);

        PVertex.x := 0;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := SIZE * 0.5;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridF;
        Inc(PVertex);

        //*  Vertical

        PVertex.x := 0;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := 0;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);

        PVertex.x := SIZE / 20.1;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := SIZE / 20.1;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridF;
        Inc(PVertex);

        PVertex.x := SIZE / 10.1;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := SIZE / 10.1;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := GRID_COLORL;
        Inc(PVertex);

        PVertex.x := SIZE / 4.05;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := SIZE / 4.05;
        PVertex.y := 0.9;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridF;
        Inc(PVertex);

        PVertex.x := SIZE / 2.025;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := SIZE / 2.025;
        PVertex.y := 0.6;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridF;
        Inc(PVertex);

        PVertex.x := SIZE / 1.352;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridD;
        Inc(PVertex);
        PVertex.x := SIZE / 1.352;
        PVertex.y := 0.4;
        PVertex.z := DEPTH;
        PVertex.color := ColorGridF;

    except
        if NOT bSuppressErrors
            then MessageDlg('Error: Exception: Init Grid vertexes.', mtWarning, [mbOk], 0);
    end;
    VertexBufferSize := GRID_VERTEX_COUNT * SizeOf(TCustomVertex);
    if Failed(d3dDevice.CreateVertexBuffer(VertexBufferSize, D3DUSAGE_WRITEONLY, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, d3dVertexBuffer, nil))
        then Exit;
    if Failed(d3dVertexBuffer.Lock(0, VertexBufferSize, PVertices, 0))
        then Exit;
    CopyMemory(PVertices, PVertexes, VertexBufferSize);
    d3dVertexBuffer.Unlock;
    Dispose(PVertexes);
    Result:= S_OK;
end;

function TSpectrum3D.InitGridGeometrySolid(d3dDevice: IDirect3DDevice9; var d3dVertexBuffer: IDirect3DVertexBuffer9): HRESULT;
const
    GRID_COLOR = $ff303030;
    DEPTH = 0;
    BOTTOM = 0;
    SIZE = 1727;
var
    PVertices: Pointer;
    PVertex: ^TCustomVertex;
    PVertexes: Pointer;
    VertexBufferSize: Cardinal;
    ColorGrid: Cardinal;
begin
    Result:= E_FAIL;
    try
        GetMem(PVertexes, GRID_VERTEX_COUNT * SizeOf(TCustomVertex));
    except
        if NOT bSuppressErrors
            then MessageDlg('Error: GetMem: ' + IntToStr(GRID_VERTEX_COUNT * SizeOf(TCustomVertex)), mtWarning, [mbOk], 0);
        Exit;
    end;
    ColorGrid := GRID_COLOR;
    PVertex := PVertexes;
    try
        PVertex.x := 0;
        PVertex.y := 0;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE * 0.985;
        PVertex.y := 0;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        PVertex.x := 0;
        PVertex.y := 1 - 1 / 1.083;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE * 0.985;
        PVertex.y := 1 - 1 / 1.083;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        PVertex.x := 0;
        PVertex.y := 1 - 1 / 1.225;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE * 0.985;
        PVertex.y := 1 - 1 / 1.225;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        PVertex.x := 0;
        PVertex.y := 1 - 1 / 1.525;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE * 0.985;
        PVertex.y := 1 - 1 / 1.525;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        PVertex.x := 0;
        PVertex.y := 1 - 1 / 2.25;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE * 0.985;
        PVertex.y := 1 - 1 / 2.25;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        PVertex.x := 0;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE * 0.985;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        //*  Vertical

        PVertex.x := 0;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := 0;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        PVertex.x := SIZE / 20.1;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE / 20.1;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        PVertex.x := SIZE / 10.1;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE / 10.1;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        PVertex.x := SIZE / 4.05;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE / 4.05;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        PVertex.x := SIZE / 2.025;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE / 2.025;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        PVertex.x := SIZE / 1.352;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE / 1.352;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);

        PVertex.x := SIZE * 0.985;
        PVertex.y := BOTTOM;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;
        Inc(PVertex);
        PVertex.x := SIZE * 0.985;
        PVertex.y := 1;
        PVertex.z := DEPTH;
        PVertex.color := ColorGrid;

    except
        if NOT bSuppressErrors
            then MessageDlg('Error: Exception: Init Grid vertexes.', mtWarning, [mbOk], 0);
    end;
    VertexBufferSize := GRID_VERTEX_COUNT * SizeOf(TCustomVertex);
    if Failed(d3dDevice.CreateVertexBuffer(VertexBufferSize, D3DUSAGE_WRITEONLY, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, d3dVertexBuffer, nil))
        then Exit;
    if Failed(d3dVertexBuffer.Lock(0, VertexBufferSize, PVertices, 0))
        then Exit;
    CopyMemory(PVertices, PVertexes, VertexBufferSize);
    d3dVertexBuffer.Unlock;
    Dispose(PVertexes);
    Result:= S_OK;
end;

function TSpectrum3D.InitGeometry(d3dDevice: IDirect3DDevice9; var d3dVertexBuffer: IDirect3DVertexBuffer9; Channel: Cardinal): HRESULT;
const
    FFT_SIZE = 2048;
var
    PVertices: Pointer;
    PVertex: ^TCustomVertex;
    i: Integer;
    FFT: Array[0..FFT_SIZE] of Single;
    Color: Cardinal;
begin
    try
        {$IFDEF USEBASSWASAPI}
        if (Channel < 255)
        AND (Channel > 0)
        then begin
            if (NOT BASS_WASAPI_SetDevice(Channel - 1))
            OR (BASS_WASAPI_GetData(@FFT, FFTSize) = DW_ERROR)
            then begin
                for i := 0 to FFT_SIZE
                    do FFT[i] := 0;
            end;
        end else begin
        {$ENDIF}
            if Bass_ChannelIsActive(Channel) <> BASS_ACTIVE_PLAYING then begin
                for i := 0 to FFT_SIZE
                    do FFT[i] := 0;
            end else BASS_ChannelGetData(Channel, @FFT, FFTSize);
        {$IFDEF USEBASSWASAPI}
        end;
        {$ENDIF}

    except
        if NOT bSuppressErrors
                then Showmessage('Error: Exception: BASS_ChannelGetData()');
    end;
    Result:= E_FAIL;
    PVertex := PFFTVertexes;
    for i := 0 to SIZE - 1 do begin        // (i - (SIZE / 2)) + 10;
        try
            PVertex.x := i * 0.9175;
            PVertex.y := SQRT(SQRT(fft[i])) * (GFXAmplitude * 140 / 128);
            PVertex.z := 0;
            Color := Trunc(255 * PVertex.y);
            PVertex.color := $ffffff00 - (Color SHL 8);
            Inc(PVertex);
        except
            if NOT bSuppressErrors
                then Showmessage('Error: Exception: Create FFT vertexes');
        end;
    end;
    if Failed(d3dDevice.CreateVertexBuffer(SIZE*SizeOf(TCustomVertex), D3DUSAGE_WRITEONLY, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, d3dVertexBuffer, nil)) then begin
        if NOT bSuppressErrors
                then Showmessage('Error: CreateVertexBuffer()');
        Exit;
    end;
    if Failed(d3dVertexBuffer.Lock(0, SIZE * SizeOf(TCustomVertex), PVertices, 0)) then begin
        if NOT bSuppressErrors
                then Showmessage('Error: LockVertexBuffer()');
        Exit;
    end;
    try
        CopyMemory(PVertices, PFFTVertexes, SIZE * SizeOf(TCustomVertex));
        d3dVertexBuffer.Unlock;
    except
        if NOT bSuppressErrors
                then Showmessage('Error: FillVertexBuffer()');
    end;
    Result := S_OK;
end;

procedure TSpectrum3D.SetupCameraMatrix(d3dDevice: IDirect3DDevice9);

    function Vector3GetMagnitude(V: TD3DVector): Double;
    begin
        Result := SQRT(V.x * V.x + V.y * V.y + V.z * V.z);
    end;

    function Vector3Normalize(V: TD3DVector): TD3DVector;
    var
        _Length: Double;
    begin
        _Length := Vector3GetMagnitude(V);
        Result.x := V.x / _Length;
        Result.y := V.y / _Length;
        Result.z := V.z / _Length;
    end;

var
    EyePointVector, LookAtPointVector, UpVector, RightVector: TD3DVector;
    ViewMatrix: TD3DMatrix;
    ProjectionMatrix: TD3DMatrix;
    ZAxis: TD3DVector;
    XAxis: TD3DVector;
    YAxis: TD3DVector;
    ScrollPos: Double;
begin
    D3DXMatrixIdentity(matWorld);
    D3DXMatrixScaling(matWorld, 0.00695 * (GFXZoom / 100), 11, 1);
    d3dDevice.SetTransform(D3DTS_WORLD, matWorld);

    ScrollPos := 6 + ((48 * (GFXZoom - 100) / (500 - 100)) * GFXScroll / 1000);  // -15

    EyePointVector := D3DXVector3(ScrollPos, 5.8, -15.0);
    LookAtPointVector := D3DXVector3(ScrollPos, 5.8, 0.0);
    UpVector := D3DXVector3(0.0, 1.0, 0.0);

    RightVector := D3DXVector3(1.0, 0.0, 0.0);
    D3DXVec3Subtract(ZAxis, LookAtPointVector, EyePointVector);
    ZAxis := Vector3Normalize(ZAxis);
    D3DXVec3Cross(XAxis, UpVector, ZAxis);
    XAxis := Vector3Normalize(XAxis);
    D3DXVec3Cross(YAxis, ZAxis, XAxis);

    ViewMatrix._11 := XAxis.x; ViewMatrix._12 := YAxis.x; ViewMatrix._13 := ZAxis.x; ViewMatrix._14 := 0.0;
    ViewMatrix._21 := XAxis.y; ViewMatrix._22 := YAxis.y; ViewMatrix._23 := ZAxis.y; ViewMatrix._24 := 0.0;
    ViewMatrix._31 := XAxis.z; ViewMatrix._32 := YAxis.z; ViewMatrix._33 := ZAxis.z; ViewMatrix._34 := 0.0;

    ViewMatrix._41 := - D3DXVec3Dot( XAxis, EyePointVector );
    ViewMatrix._42 := - D3DXVec3Dot( YAxis, EyePointVector );
    ViewMatrix._43 := - D3DXVec3Dot( ZAxis, EyePointVector );
    ViewMatrix._44 := 1.0;

    //D3DXMatrixLookAtLH(matView, vEyePt, vLookatPt, vUpVec);
    d3dDevice.SetTransform(D3DTS_VIEW, ViewMatrix);

    // For the projection matrix, we set up a perspective transform (which
    // transforms geometry from 3D view space to 2D viewport space, with
    // a perspective divide making objects smaller in the distance). To build
    // a perpsective transform, we need the field of view (1/4 pi is common),
    // the aspect ratio, and the near and far clipping planes (which define at
    // what distances geometry should be no longer be rendered).
    D3DXMatrixPerspectiveFovLH(ProjectionMatrix, D3DX_PI/4, 1.0, 1.0, 50000);
    d3dDevice.SetTransform(D3DTS_PROJECTION, ProjectionMatrix);
end;

end.
